*****************************************************************************
** Engine: 14 Bit Amplifier II                                             **
**                                                                         **
** Project: Eagleplayer 2.05                                               **
** Authors: Jan Blumenthal & Henryk Richter                                **
** Start  : 1993/01/09                                                     **
** $Rev: 26 $ $Author: buggs $                                             **
** $Date: 2016-08-22 20:24:07 +0200 (Mo, 22 Aug 2016) $                   **
**                                                                         **
*****************************************************************************
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program (See the included file COPYING);
** if not, write to the Free Software Foundation, Inc.,
** 675 Mass Ave, Cambridge, MA 02139, USA.
**
**************************************************************************** 
; Input:
;        A4   - AudioStruct for current channel and panning
;        A2   - Output buffer for "SamLEN*2" shorts
;        D0   - number of samples (SamLEN*2)
;        A6   - Volume table
; Macro Arguments:
; - bailout pointer (return without doing anything)
; - volume table adjust (0=no -> current 16 Bit mixing routine)
; - double offsets (1) or single byte offsets (0), double offsets used in 16 bit mixing
; Output:
;        D0   - current offset towards A0 (upper word is current fractional position)
;        D4   - interpolation step size, lower word integer part, upper word fractional part
;        D6   - number of samples pending in current instrument
;        A0   - current source sample, offset by (already) played bytes
;        A1   - 16*D4+1, threshold for skipping the 16 sample "fast" loop
;        A5   - remaining samples in current instrument
;        A6   - Volume Table position for current volume
MIX_INIT	macro
	move	d0,a5		; number of copied samples (not -16 here, because of the branches to buffer fill) 

	tst.b	AS_DmaBit(a4)
	beq	\1		; DMA disable, nothing playing

	tst.b	AS_NoLoop(a4)
	bne	\1		; noloop, sample done

	ifne	\2		
	move	AS_LeftVolume(a4),d0	; volume table 2b changed for panning support
	lsl.w	#8,d0			; 8 Bit: 256 entries in volume table (2 bytes per entry)
	lea	(A6,D0.w*2),A6		; position in volume table
	endc

	tst.l	AS_Currentadr(a4)	; check sample address and reload if necessary
	bne.s	.mixm_cont		; prepare new sample to be played, return here
					; Note: branches out if no sample

;-------- copy new sample address into active fields and reset counter
; Input:  A4- current AS struct
; Output: no registers modified
	clr.l	AS_CurrentPos(a4)			; will be overridden in preload function
	move.l	AS_RepeatAdr(A4),AS_CurrentAdr(a4)
	beq	\1			; we would need repeat right now but repeat address is 0, stop playing
	move.l	AS_RepeatSize(a4),AS_SampleSize(a4)
	beq	\1			; same for repeat size
	bra.s	.mixm_addressdone

;-------- compute interpolation step size from Paula constant, current sampling frequency and sample period
; Input:  D0 = current period, A4=current AS struct
; Output: D4 = new interpolation step size
.mixm_set_interpolation:
	move.w	d0,ASP_LastPeriod(a4)
	beq	\1		; no sample, return
	; precompute interpolation factor and just compare AS_Period with ASPRIV_Period
	ifne	1
		; divu.l requires 68020+ -> this one is more accurate and faster at the same time
		move.l	referenceperiod(pc),d4	;
		ext.l	d0			;just to be sure (sample_period*4)
		divu.l	d0,d4			;mixing_period/sample_period*65536
		;subq.l	#1,d4			;workaround for old fm_mix which previously required that the integer step size is zero
		swap	d4			
	else
		; 68000 compatible routine
		move.l	clock(pc),d4		;reference clock: PAL timing
		divu	d0,d4			;/period (*4) -> 0.25*mixrate in d4
		swap	d4			;0.25*mixrate * 65536
		clr	d4			;clear lower word
		lsr.l	#2,d4			;mixrate * 16384
		move.l	mrate(pc),d0		;
		divu	d0,d4
		swap	d4
		clr	d4
		rol.l	#4,d4			;
	endc
	move.l	d4,ASP_LastInterpolate(a4)
	ifne	\3
		move.l	d4,d0
		add.l	#1<<(31-FM_tabs_Bits),d0	;round to nearest...
		clr.w	d0
		rol.l	#FM_tabs_Bits,d0
		lea	FM_tabs(pc),a3
		move.l	(a3,d0.w*4),d0
		;move.l	#FM_tab_175,d0			;test only
		move.l	d0,FM_tab
	endc
	bra.s	.mixm_interpolation_done

.mixm_addressdone:
.mixm_cont:

	; check period and precomputed period
	move.w	AS_Period(a4),d0	   ;current period
	move.l	ASP_LastInterpolate(a4),d4 ;prefetch interpolation step size

	cmp.w	ASP_LastPeriod(a4),d0		; period still valid ?
	bne.s	.mixm_set_interpolation	; Note: branches out if period = 0
.mixm_interpolation_done:
	;
	move.l	d4,d0			;counter for nearest neighbor
	swap	d0			;lower word: fractional part
	lsl.l	#4,d0			;*16 = 16 samples mixed in one run
	swap	d0			;Overflow count, also wieviele neue Bytes zu holen sind
	addq.w	#2,d0			;this +2 is inaccurate but leads to one less instruction in main loop
	move.l	d0,a1		;-> store overflow count in A1

	; sample address, current sample position and fractional position
	move.l	AS_Currentadr(a4),a0	; pos (addr of current instrument)
	move.l	AS_CurrentPos(a4),d6	; position within sample
	add.l	d6,a0			; current position (=first unplayed sample) 
	ifne	\3
	add.l	d6,a0
	endc

	moveq	#0,d0
	move.w	AS_CurrentFPOS(a4),d0	;fpos (fractional position) in lower word
	swap	d0			;clear lower word

	endm


; limit D6 to 16 bit unsigned, if sounds longer than 32k are in use 
; (we need just a fraction of the input samples per run, so clamping is an easy workaround)
; second argument: counter or extra identifier
MIX_LIMITER	macro
        cmp.l   #$7fff,\1               ;for "just" started or repeated samples, clamp size
        bls.s   .mixm_SamLEN_ok\2
        move.l  #$7fff,\1               ;in best case, done only once per run, so don`t optimize constant into register
.mixm_SamLEN_ok\2:
		endm

MIX_REMBYTES	macro

        neg.l   \1                      ;current position (played length)
	ifeq	\3
       	add.l   AS_SampleSize(a4),\1    ;total length - played length = remaining bytes
	else
        move.l  AS_SampleSize(a4),\2    ;total length in bytes/2 - played length in words = remaining words
        lsr.l   #1,\2
        add.l   \2,\1
	endc

		endm
; MIX_REMBYTES d6,d1,0 (8 Bit samples), d6=current position, d1=scratch register
; MIX_REMBYTES d6,d1,1 (16 Bit samples)



MIX_LOOPCHECK	macro
	st	AS_NoLoop(a4)			;stop playing current instrument (preliminary)

	move.b	AS_LoopFlag(a4),d1
	beq.w	\1				;no loop, just set the rest to zero (or leave in case of mix2 code)

	sf	AS_NoLoop(a4)			;ok, we are looping

	btst	#ASLoopF_PingPong,d1		;branch to Ping-Pong-Loop routine (PC tunes)
	beq.s	.mixm_lc_NormalLoop

	btst	#ASLoopF_Looping,AS_CurrentFPOS+2(a4) ; ping-pong: was forward looping ?
	bne.s	\2				; yes, so go backwards

.mixm_lc_NormalLoop:
	bset	#ASLoopF_Looping,AS_CurrentFPOS+2(a4) ; we are looping forward

	move.l	AS_RepeatAdr(a4),a0		;loop start
	move.l	a0,AS_CurrentAdr(a4)
	moveq	#0,d1
	move.w	d0,d1				;currently played length (keep track here due to mixing stride, pointless with oversampling btw.)
	add.l	AS_CurrentPos(a4),d1

	ifne	\3
		move.l	AS_Samplesize(A4),d6
		lsr.l	#1,d6
		sub.l	d6,d1
;		bge.s	.mixm_lc_subok
;		moveq	#0,d1
;.mixm_lc_subok:

	else
		sub.l	AS_Samplesize(A4),d1
	endc
;	add.l	#1,d1

	clr.l	AS_CurrentPos(a4)
;	move.l	d1,AS_CurrentPos(a4)		; reset position
	move.w	d1,d0				; integer position

	move.l	AS_RepeatSize(a4),d6		;loop len
	move.l	d6,AS_SampleSize(a4)
	ifne	\3
		lsr.l	#1,d6
	endc
;	sub.l	d1,d6				;Length = remaining bytes (or words)

	tst.l	AS_Int(a4)			;"Interrupt" bei neuem Sample ?
	beq.s	.mixm_lc_noInt1
	bsr	mix_int
.mixm_lc_noINT1:

	endm

;MIX_LOOPCHECK	.fm_mix_clear_rest,.fm_mix_PingPongLoop	;1st arg: bra target for one-shot samples, 2nd arg: bra target for ping-pong loop code

MIX_POSTCHECK	macro
	moveq	#0,d6
	move	d0,d6
	add.l	AS_CurrentPos(a4),d6
	move.l	d6,AS_CurrentPos(a4)

	swap	d0
	move.w	d0,AS_CurrentFPOS(a4)	;Fraction Pos

	move.l	AS_SampleSize(A4),d1
	ifne	\2
		lsr.l	#1,d1
	endc

	sub.l	d1,d6				;current position beyond end of instrument ? (remember: d6 loaded from d0 and incremented)
	blt.s	\1

	move.b	AS_LoopFlag(a4),d1
	bne.b	.fm_mixm_yes_loop
	
	st	AS_NoLoop(a4)			; sample is not looping
	bra.b	\1

	; condition: end of sample and looping is on
.fm_mixm_yes_loop:
	bset	#ASLoopF_Looping,AS_CurrentFPOS+2(a4)

	move.l	AS_RepeatSize(a4),a0		;loop len
	move.l	a0,AS_SampleSize(a4)
	move.l	AS_Repeatadr(a4),a0
	move.l	A0,AS_CurrentAdr(a4)

	move.l	d6,AS_CurrentPos(a4)

	tst.l	AS_Int(a4)		;"Interrupt" bei neuem Sample ?
	beq.s	\1
	bsr	mix_int		;needs A4
	endm
;	MIX_POSTCHECK	.fm_mix_bfull_ret,0 ;where 0/1 = byte samples (8 Bit) or word samples (16 Bit)

; Initialize Ping-pong loop: copy parameters for backward looping and conditionally
; issue interrupt at restart
; D0 - number of played samples in current run
; D4 - interpolation step size (fractional<<16 + integer)
; A0 - input samples (input instrument)
; A2 - output samples (mixing buffer)
; A3 - interpolation table (for fm_mix variants)
; A4 - current AS Struct
; A5 - remaining samples in output
; A6 - volume table
MIX_PINGPONG_INIT	macro

        bset    #ASLoopF_BackLooping,AS_CurrentFPOS+2(a4) ; now loop backwards

        move.l  AS_RepeatAdr(a4),a0             ;loop start
        move.l  a0,AS_CurrentAdr(a4)

        moveq   #0,d1
        move.w  d0,d1                           ;im aktuellen Rundlauf gespielte Laenge
        add.l   AS_CurrentPos(a4),d1
        sub.l   AS_Samplesize(A4),d1            ;?bertrag ?ber eigentliche aktuelle Laenge

        move.l  AS_Repeatsize(a4),d6
        move.l  d6,AS_Samplesize(a4)
        sub.l   d1,d6
        subq.l  #1,d6
        move.l  d6,AS_CurrentPos(a4)            ;aktuelle Position = Repeatsize-1-Uebertrag
        add.l   d6,a0                           ;Aktuelle Position
        clr     d0                              ;reset integer part

        tst.l   AS_Int(a4)              ;"Interrupt" bei neuem Sample ?
        beq.s   .mixm_noInt_P
        bsr     mix_int
.mixm_noINT_P:
        move.w  A5,d1                   ;still samples left
        bgt     \1                      ;yep, do loop backwards

			endm

; the filtered mixing performs a lookahead
MIX_FM_FADEOUT_TIME	EQU	4	; 4 bytes fadeout period (counted at input sampling rate)
MIX_FM_FADEOUT	macro

        ; non-looping instruments require some extra treatment:
        ;  - we have a look-ahead in here, so that 2 "future" samples are already stored
        ;  - provide a "dummy" extension with "0"-samples
        ;  - no problem with "repeating" instruments since the integer position is reset and
        ;    new samples are loaded as needed
        move.b  AS_LoopFlag(a4),d1              ;no loop ?
        bne.s   .mixm_loop_considerations

        st      AS_NoLoop(a4)                   ;stop playing current sample (preliminary)
        lea     SilenceBuffer,a0		;ZeroSamplebuffer,A0
        cmpa.l  AS_CurrentAdr(A4),A0            ;
        beq.w 	\1 

        sf      AS_NoLoop(a4)                   ;allow continue to play and use dummy samples
        move.l  A0,AS_CurrentAdr(A4)
        moveq   #MIX_FM_FADEOUT_TIME,d6		;
        move.l  d6,AS_SampleSize(A4)            ;fade out (two would be sufficient)
        clr.l   AS_CurrentPos(A4)
        clr.w   D0
        bra.w   \2				;try again

.mixm_loop_considerations:

		endm
;	MIX_FM_FADEOUT .fm_mix_clear_rest,.fm_mix_mainloop



; backward looping done, now go forward
; no arguments (yet)
MIX_PING_DIRECTIONCHANGE	macro

		neg.w	d0		;|played length| (for carry, relevant only in "subsampling" case )

		bclr	#ASLoopF_BackLooping,AS_CurrentFPOS+2(a4) ; forward

		move.l	AS_RepeatAdr(a4),a0		;loop start
		move.l	a0,AS_CurrentAdr(a4)

		moveq	#0,d1
		move.w	d0,d1				;im aktuellen Rundlauf gespielte Laenge
		sub.l	AS_CurrentPos(a4),d1
		move.l	d1,AS_CurrentPos(a4)
		add.l	d1,a0

		move.l	AS_RepeatSize(a4),d6		;loop len
		move.l	d6,AS_SampleSize(a4)
		sub.l	d1,d6				;Len - Current Position
		clr	d0				;reset integer part

		tst.l	AS_Int(a4)		;"Interrupt" bei neuem Sample ?
		beq.s	.mixm_PingB_noInt
		bsr	mix_int		;needs A4
.mixm_PingB_noINT:
					;in D6 uebrige Laenge (currentPos = Rest vom aktuellen
					;Wert bis 0)
		move.l	a1,d1
		neg.w	d1
		move.l	d1,a1

				endm

; check whether the instrument ended at buffer full condition (in backward pingpong mode)
;
MIX_PING_BUFFERFULL	macro

		neg.w	d0		;|gespielte Laenge| (fuer Uebertrag)
		moveq	#0,d1
		move	d0,d1
		swap	d0
		move.w	d0,AS_CurrentFPOS(a4)	;Fraction Pos

		move.l	AS_CurrentPos(a4),d6
		sub.l	d1,d6
		move.l	d6,AS_CurrentPos(a4)
		bgt.s	\1	

		bclr	#ASLoopF_BackLooping,AS_CurrentFPOS+2(a4)

		move.l	AS_RepeatAdr(a4),a0		;loop start
		move.l	a0,AS_CurrentAdr(a4)

		neg.l	d6
		move.l	d6,AS_CurrentPos(A4)

		move.l	AS_RepeatSize(a4),d6		;loop len
		move.l	d6,AS_SampleSize(a4)

		tst.l	AS_Int(a4)		;"Interrupt" bei neuem Sample ?
		beq.s	\1
		bsr	mix_int		; start interrupt

			endm
; used:
;  A0 - input samples
;  D0 - offset to a0 (.w), fractional position (upper word)
;  D4 - increment
;  A2 - output
;  A3 - interpolation table
;  A6 - volume table
; Trashed:
;  D1-d3,d5,d7
;
; 4 tables (shuffled to match the ringbuffer)
; 16 (see FM_Accuracy) steps, addressed by upper bits of d4
; 4  (see FM_Taps) samples in ringbuffer which require a multiplication constant
; Ringbuffer states: (example starts with "current" sample 0, smaller is past, bigger is future)
; AS_FMPos=3
; (0) 1  2  -1
; AS_FMPos=0
;  0 (1) 2  3
; AS_FMPos=1
;  4  1 (2) 3
; AS_FMPos=2
;  4  5  2 (3)
fm_mix_one_ac:	macro	;anticlick sinc mixer, copy phase
		;moveq	#0,d3
		move.l	d0,d3				;upper 16 bit: fractional position
		rol.l	#FMB_Accuracy+FMB_Taps+1,d3	;
		and.l	#(FM_Accuracy*FM_Taps*2)-(1<<(FMB_Taps+1)),d3	;currently relevant 4 coefficients
		movem.w	AS_FMSamples1(A4),d1-d2/d5/d7
		muls.w	 (a3,d3.w),d1
		muls.w	2(a3,d3.w),d2
		muls.w	4(a3,d3.w),d5
		add.l	d2,d1
		muls.w	6(a3,d3.w),d7
		add.l	d5,d1
		add.l	d7,d1				;sum over all 4 coefficients * (past+future samples)
		add.l	d1,d1				;coefficients are $7fff max. -> double the sum
		swap	d1				;get upper half
		move.w	d1,(a2)+
		add.l	d4,d0
		bcc.s	nooverflac1\1$
	ifne    anticlick
		; load last sample from buffer and apply IIR filter
		moveq	#-(AC_DELAY),d3			;"current" sample in ringbuffer
		add.b	AS_FMPos(a4),d3			;ringbuffer index
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.w	AS_FMSamples1(A4,d3.w*2),d1	;y[n-1]
		addq.b	#AC_DELAY,d3			;
		and.w	#FM_Taps-1,d3			;revert ringbuffer index to new write position
		move.w	ASP_Anticlickpos(a4),d5		;
		muls.w	d5,d1				;y[n-1] * anticlickpos
		sub.w   #AC_REF,d5			;anticlickpos-64

		moveq	#0,d2
		move.b	(A0,D0.w),d2			;get sample
		addq.w	#1,d0				;D0 post increment
	 	move.w	(A6,d2.w*2),d2			;volume correct
		muls.w	d5,d2				;x[n] * (anticlickpos-64)
		sub.l	d2,d1				;y[n]
		asr.l	#AC_SHIFT,d1			;

		move.w	d1,AS_FMSamples1(A4,d3.w*2)	;
		addq.w	#1,d3
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.b	d3,AS_FMPos(A4)
		lsl.w	#FMB_Accuracy+FMB_Taps+1,d3	;accuracy bits + taps bits + word size
		lea	FM_tab_100(pc),a3		;table for current ringbuffer position
		ext.l	d3
		add.l	d3,a3

		add.w   #AC_REF,d5
		moveq	#0,d2
		subq.w	#1,d5
		addx.w	d2,d5
		move.w	d5,ASP_Anticlickpos(a4)		;
	else
		;addq.w	#1,d0				;D0 increment (deprecated)
		moveq	#0,d2
		move.b	(A0,D0.w),d2			;get sample
		addq.w	#1,d0				;D0 post increment
	 	move.w	(A6,d2.w*2),d2			;volume correct
		moveq	#0,d3
		move.b	AS_FMPos(a4),d3			;ringbuffer index
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.w	d2,AS_FMSamples1(A4,d3.w*2)	;
		addq.w	#1,d3
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.b	d3,AS_FMPos(A4)
		lsl.w	#FMB_Accuracy+FMB_Taps+1,d3	;accuracy bits + taps bits + word size
		lea	FM_tab_100(pc),a3		;table for current ringbuffer position
		add.l	d3,a3
	endc
nooverflac1\1$
		endm

fm_mix_one_ac_2:	macro	;anticlick sinc mixer, add phase (behaves like normal one when anticlick=0)
		;moveq	#0,d3
		move.l	d0,d3				;upper 16 bit: fractional position
		rol.l	#FMB_Accuracy+FMB_Taps+1,d3	;
		and.l	#(FM_Accuracy*FM_Taps*2)-(1<<(FMB_Taps+1)),d3	;currently relevant 4 coefficients
		movem.w	AS_FMSamples1(A4),d1-d2/d5/d7
		muls.w	 (a3,d3.w),d1
		muls.w	2(a3,d3.w),d2
		muls.w	4(a3,d3.w),d5
		muls.w	6(a3,d3.w),d7
		add.l	d2,d1
		add.l	d5,d7
		add.l	d7,d1				;sum over all 4 coefficients * (past+future samples)
		add.l	d1,d1				;coefficients are $7fff max. -> double the sum
		swap	d1				;get upper half
		add.w	d1,(a2)+
		add.l	d4,d0
		bcc.s	nooverflac2\1$
	ifne    anticlick
		; load last sample from buffer and apply IIR filter
		moveq	#-(AC_DELAY),d3			;"current" sample in ringbuffer
		add.b	AS_FMPos(a4),d3			;ringbuffer index
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.w	AS_FMSamples1(A4,d3.w*2),d1	;y[n-1]
		addq.b	#AC_DELAY,d3			;
		and.w	#FM_Taps-1,d3			;revert ringbuffer index to new write position
		move.w	ASP_Anticlickpos(a4),d5		;
		muls.w	d5,d1				;y[n-1] * anticlickpos
		sub.w   #AC_REF,d5			;anticlickpos-64

		moveq	#0,d2
		move.b	(A0,D0.w),d2			;get sample
		addq.w	#1,d0				;D0 post increment
	 	move.w	(A6,d2.w*2),d2			;volume correct
		muls.w	d5,d2				;x[n] * (anticlickpos-64)
		sub.l	d2,d1				;y[n]
		asr.l	#AC_SHIFT,d1			;

		move.w	d1,AS_FMSamples1(A4,d3.w*2)	;
		addq.w	#1,d3
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.b	d3,AS_FMPos(A4)
		lsl.w	#FMB_Accuracy+FMB_Taps+1,d3	;accuracy bits + taps bits + word size
		ext.l	d3
		lea	FM_tab_100(pc),a3		;table for current ringbuffer position
		add.l	d3,a3

		add.w   #AC_REF,d5
		moveq	#0,d2
		subq.w	#1,d5
		addx.w	d2,d5
		move.w	d5,ASP_Anticlickpos(a4)		;
	else
		;addq.w	#1,d0				;D0 increment (deprecated)
		moveq	#0,d2
		move.b	(A0,D0.w),d2			;get sample
		addq.w	#1,d0				;D0 post increment
	 	move.w	(A6,d2.w*2),d2			;volume correct
		moveq	#0,d3
		move.b	AS_FMPos(a4),d3			;ringbuffer index
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.w	d2,AS_FMSamples1(A4,d3.w*2)	;
		addq.w	#1,d3
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.b	d3,AS_FMPos(A4)
		lsl.w	#FMB_Accuracy+FMB_Taps+1,d3	;accuracy bits + taps bits + word size
		lea	FM_tab_100(pc),a3		;table for current ringbuffer position
		add.l	d3,a3
	endc
nooverflac2\1$
		endm

fm_mix_one:	macro
		;moveq	#0,d3
		move.l	d0,d3				;upper 16 bit: fractional position
		rol.l	#FMB_Accuracy+FMB_Taps+1,d3	;
		and.l	#(FM_Accuracy*FM_Taps*2)-(1<<(FMB_Taps+1)),d3	;currently relevant 4 coefficients
		movem.w	AS_FMSamples1(A4),d1-d2/d5/d7
		muls.w	 (a3,d3.w),d1
		muls.w	2(a3,d3.w),d2
		muls.w	4(a3,d3.w),d5
		muls.w	6(a3,d3.w),d7
		add.l	d2,d1
		add.l	d5,d7
		add.l	d7,d1				;sum over all 4 coefficients * (past+future samples)
		add.l	d1,d1				;coefficients are $7fff max. -> double the sum
		swap	d1				;get upper half
		move.w	d1,(a2)+
		add.l	d4,d0
		bcc.s	nooverfl\1$
		moveq	#0,d2
		move.b	(A0,D0.w),d2			;get sample
		addq.w	#1,d0				;D0 post increment
	 	move.w	(A6,d2.w*2),d2			;volume correct
		moveq	#0,d3
		move.b	AS_FMPos(a4),d3			;ringbuffer index
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.w	d2,AS_FMSamples1(A4,d3.w*2)	;
		addq.w	#1,d3
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.b	d3,AS_FMPos(A4)
		lsl.w	#FMB_Accuracy+FMB_Taps+1,d3	;accuracy bits + taps bits + word size
		lea	FM_tab_100(pc),a3		;table for current ringbuffer position
		add.l	d3,a3
nooverfl\1$
		endm

fm_mix_one_2:	macro
		;moveq	#0,d3
		move.l	d0,d3				;upper 16 bit: fractional position
		rol.l	#FMB_Accuracy+FMB_Taps+1,d3	;
		and.l	#(FM_Accuracy*FM_Taps*2)-(1<<(FMB_Taps+1)),d3	;currently relevant 4 coefficients
		movem.w	AS_FMSamples1(A4),d1-d2/d5/d7
		muls.w	 (a3,d3.w),d1
		muls.w	2(a3,d3.w),d2
		muls.w	4(a3,d3.w),d5
		muls.w	6(a3,d3.w),d7
		add.l	d2,d1
		add.l	d5,d7
		add.l	d7,d1				;sum over all 4 coefficients * (past+future samples)
		add.l	d1,d1				;coefficients are $7fff max. -> double the sum
		swap	d1				;get upper half
		add.w	d1,(a2)+
		add.l	d4,d0
		bcc.s	nooverfl\1$
		moveq	#0,d2
		move.b	(A0,D0.w),d2			;get sample
		addq.w	#1,d0				;D0 post increment
	 	move.w	(A6,d2.w*2),d2			;volume correct
		moveq	#0,d3
		move.b	AS_FMPos(a4),d3			;ringbuffer index
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.w	d2,AS_FMSamples1(A4,d3.w*2)	;
		addq.w	#1,d3
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.b	d3,AS_FMPos(A4)
		lsl.w	#FMB_Accuracy+FMB_Taps+1,d3	;accuracy bits + taps bits + word size
		lea	FM_tab_100(pc),a3		;table for current ringbuffer position
		add.l	d3,a3
nooverfl\1$
		endm


; used:
;  A0 - input samples
;  D0 - offset to a0 (.w), fractional position (upper word)
;  D4 - increment
;  A2 - output
;  A3 - interpolation table
;  A6 - volume table
; Trashed:
;  D1-d3,d5,d7
;
; 4 tables (shuffled to match the ringbuffer)
; 16 (see FM_Accuracy) steps, addressed by upper bits of d4
; 4  (see FM_Taps) samples in ringbuffer which require a multiplication constant
; Ringbuffer states: (example starts with "current" sample 0, smaller is past, bigger is future)
; AS_FMPos=3
; (0) 1  2  -1
; AS_FMPos=0
;  0 (1) 2  3
; AS_FMPos=1
;  4  1 (2) 3
; AS_FMPos=2
;  4  5  2 (3)
fm16_mix_one:	macro
		;moveq	#0,d3
		move.l	d0,d3				;upper 16 bit: fractional position
		rol.l	#FMB_Accuracy+FMB_Taps+1,d3	;
		and.l	#(FM_Accuracy*FM_Taps*2)-(1<<(FMB_Taps+1)),d3	;currently relevant 4 coefficients
		movem.w	AS_FMSamples1(A4),d1-d2/d5/d7
		muls.w	 (a3,d3.w),d1
		muls.w	2(a3,d3.w),d2
		muls.w	4(a3,d3.w),d5
		muls.w	6(a3,d3.w),d7
		add.l	d2,d1
		add.l	d5,d7
		add.l	d7,d1				;sum over all 4 coefficients * (past+future samples)
		add.l	d1,d1				;coefficients are $7fff max. -> double the sum
		swap	d1				;get upper half
		move.w	d1,(a2)+
		add.l	d4,d0
		bcc.s	nooverfl\1$
		;addq.w	#1,d0				;D0 increment (deprecated)
		moveq	#0,d2
		move.w	(A0,D0.w*2),d2			;get sample
		addq.w	#1,d0				;D0 post increment
	 	move.w	(A6,d2.l*2),d2			;volume correct
		moveq	#0,d3
		move.b	AS_FMPos(a4),d3			;ringbuffer index
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.w	d2,AS_FMSamples1(A4,d3.w*2)	;
		addq.w	#1,d3
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.b	d3,AS_FMPos(A4)
		lsl.w	#FMB_Accuracy+FMB_Taps+1,d3	;accuracy bits + taps bits + word size
		lea	FM_tab_100(pc),a3		;table for current ringbuffer position
		add.l	d3,a3
nooverfl\1$
		endm

; used:
;  A0 - input samples
;  D0 - offset to a0 (.w), fractional position (upper word)
;  D4 - increment
;  A2 - output
;  A3 - interpolation table
;  A6 - volume table
; Trashed:
;  D1-d3,d5,d7
;
; 4 tables (shuffled to match the ringbuffer)
; 16 (see FM_Accuracy) steps, addressed by upper bits of d4
; 4  (see FM_Taps) samples in ringbuffer which require a multiplication constant
; Ringbuffer states: (example starts with "current" sample 0, smaller is past, bigger is future)
; AS_FMPos=3
; (0) 1  2  -1
; AS_FMPos=0
;  0 (1) 2  3
; AS_FMPos=1
;  4  1 (2) 3
; AS_FMPos=2
;  4  5  2 (3)
fm16down_mix_one:	macro
		;moveq	#0,d3
		move.l	d0,d3				;upper 16 bit: fractional position
		rol.l	#FMB_Accuracy+FMB_Taps+1,d3	;
		and.l	#(FM_Accuracy*FM_Taps*2)-(1<<(FMB_Taps+1)),d3	;currently relevant 4 coefficients
		movem.w	AS_FMSamples1(A4),d1-d2/d5/d7
		muls.w	 (a3,d3.w),d1
		muls.w	2(a3,d3.w),d2
		muls.w	4(a3,d3.w),d5
		muls.w	6(a3,d3.w),d7
		add.l	d2,d1
			moveq	#0,d2
		add.l	d5,d7
			move.w	(A0,D0.w*2),d2			;get sample
			moveq	#0,d3
	 		move.w	(A6,d2.l*2),d2			;volume correct
		add.l	d7,d1				;sum over all 4 coefficients * (past+future samples)
			move.b	AS_FMPos(a4),d3			;ringbuffer index
		add.l	d1,d1				;coefficients are $7fff max. -> double the sum
			and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		swap	d1				;get upper half
			move.w	d2,AS_FMSamples1(A4,d3.w*2)	;
			addq.w	#1,d3
		move.w	d1,(a2)+
			and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
			move.b	d3,AS_FMPos(A4)
			lsl.w	#FMB_Accuracy+FMB_Taps+1,d3	;accuracy bits + taps bits + word size
			move.l	FM_tab(pc),a3			;table for current ringbuffer position
			add.l	d3,a3
skipload\1$
		;
		add.l	d4,d0
		bcc.s	nooverfl\1$
		ifne	\2
		cmp	d0,d6
		bgt.s	noendofsample\1$		;at last sample -> don't load next
		addq.w	#1,d0
		bra.s	nooverfl\1$
noendofsample\1$
		endc
		;addq.w	#1,d0				;D0 increment (deprecated)
		moveq	#0,d2
		move.w	(A0,D0.w*2),d2			;get sample
		addq.w	#1,d0				;D0 post increment
	 	move.w	(A6,d2.l*2),d2			;volume correct
		moveq	#0,d3
		move.b	AS_FMPos(a4),d3			;ringbuffer index
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.w	d2,AS_FMSamples1(A4,d3.w*2)	;
		addq.w	#1,d3
		and.w	#FM_Taps-1,d3			;make sure, the position is sane (assumes 2^n taps)
		move.b	d3,AS_FMPos(A4)
		lsl.w	#FMB_Accuracy+FMB_Taps+1,d3	;accuracy bits + taps bits + word size
		move.l	FM_tab(pc),a3			;table for current ringbuffer position
		add.l	d3,a3
nooverfl\1$
		endm


